/*
 *      signalshed.c
 *      
 *      Copyright 2010 Alex <alex@linuxonly.ru>
 *      
 *      This program is free software: you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 3 of the License, or
 *      (at your option) any later version.
 *      
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *      
 *      You should have received a copy of the GNU General Public License
 *      along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <time.h>
#include <string.h>

#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/sem.h>
#include <sys/mman.h>
#include <signal.h>

#include "signalshed.h"

#define SHMSIZE(x) ( sizeof(int)*2 + sizeof(pid_t)*(x) )
#define SEMID(x) ( ((int *)(x))[0] )
#define MAXCLIENTS(x) ( ((int *)(x))[1] )
#define CLIENTS(x, n) ( ((pid_t *)((char *)(x) + sizeof(int)*2))[(n)] )

union semun {
	int val;
	struct semid_ds *buf;
	ushort *array;
};

void signalshed_init(signalshed_t *signalshed, key_t svkey, int maxprocesses)
{
	unsigned short semvalues[1];
	union semun semarg;
	
	*signalshed = (signalshed_t)mmap(NULL, SHMSIZE(maxprocesses), PROT_READ|PROT_WRITE, MAP_SHARED|MAP_ANONYMOUS, 0, 0);
	memset(*signalshed, 0, SHMSIZE(maxprocesses));
	
	SEMID(*signalshed) = semget(svkey, 1, 0666 | IPC_CREAT);
	MAXCLIENTS(*signalshed) = maxprocesses;
	
	semvalues[0] = 0;
	
	semarg.array = (unsigned short *)&semvalues;
	if (semctl(SEMID(*signalshed), 0, SETALL, semarg) == -1) {
		perror("[SIGNAL SHEDULER] semctl call:");
	}
	
}

void signalshed_destroy(signalshed_t signalshed)
{
	int semid, clients;
	int i;
	
	/*for (i=MAXCLIENTS(signalshed); i>0; i--) {
		if ((CLIENTS(signalshed, i) > 0) && (kill(CLIENTS(signalshed, i), 0) == 0)) {
			kill(CLIENTS(signalshed, i), SIGTERM);
		}
	}*/
	
	semid = SEMID(signalshed);
	clients = MAXCLIENTS(signalshed);
	semctl(semid, 1, IPC_RMID);
	munmap(signalshed, (size_t)(SHMSIZE(clients)));
}

void signalshed_add_client(signalshed_t signalshed, pid_t pid, int seq)
{
	int i;
	
	if (seq < 0) {
		for (i=0; i<MAXCLIENTS(signalshed); i++) {
			if (CLIENTS(signalshed, i) == 0) {
				CLIENTS(signalshed, i) = pid;
			}
		}
	} else {
		CLIENTS(signalshed, seq) = pid;
	}
}

void signalshed_remove_client(signalshed_t signalshed, pid_t pid, int seq)
{
	int i;
	int res = 0;
	union semun semarg;
		
	if (seq < 0) {
		for (i=0; i<MAXCLIENTS(signalshed); i++) {
			if (CLIENTS(signalshed, i) == pid) {
				CLIENTS(signalshed, i) = 0;
				res = 1;
			}
		}
	} else {
		CLIENTS(signalshed, seq) = 0;
		res = 1;
	}
	
	if (res) {
		if (semctl(SEMID(signalshed), 0, GETZCNT, semarg) > 0) {
			semarg.val = 0;
			if (semctl(SEMID(signalshed), 0, SETVAL, semarg) == -1) {
				perror("[SIGNAL SHEDULER] semctl call:");
			}
		}
	}
}

void signalshed_wait_value(signalshed_t signalshed, int val)
{
	union semun semarg;
	
	if (val <= 0) return;
	
	semarg.val = val;
	if (semctl(SEMID(signalshed), 0, SETVAL, semarg) == -1) {
		perror("[SIGNAL SHEDULER] semctl call:");
	}
}

void signalshed_wait(signalshed_t signalshed)
{
	struct sembuf buf[1];
	struct timespec ts;
	
	ts.tv_sec = 0;
	ts.tv_nsec = 500000;
	
	buf[0].sem_num = 0;
	buf[0].sem_op = 0;
	buf[0].sem_flg = 0;
	
	#ifdef linux
		if (semtimedop(SEMID(signalshed), (struct sembuf*)&buf, 1, &ts) == -1) {
			perror("[SIGNAL SHEDULER] semop call:");
		}
	#else
		if (semop(SEMID(signalshed), (struct sembuf*)&buf, 1) == -1) {
			perror("[SIGNAL SHEDULER] semop call:");
		}
	#endif
}

void signalshed_post(signalshed_t signalshed)
{
	union semun semarg;
	
	semarg.val = 0;
	if (semctl(SEMID(signalshed), 0, SETVAL, semarg) == -1) {
		perror("[SIGNAL SHEDULER] semctl call:");
	}
}

void signalshed_continue(signalshed_t signalshed)
{
	int i;
	
	for (i=0; i<MAXCLIENTS(signalshed); i++) {
		if ((CLIENTS(signalshed, i) > 0) && (kill(CLIENTS(signalshed, i), 0) == 0)) {
			kill(CLIENTS(signalshed, i), SIGCONT);
		}
	}
}

int signalshed_main(signalshed_t signalshed, int *index, unsigned int *errorcount, unsigned int *idlecount, int (*send_cb)(int *index, unsigned int *errorcount, unsigned int *idlecount))
{
	sigset_t sigmask;
	int signum;
	
	if ((send_cb == NULL) || (*index <= 0)) return 1;
	
	sigemptyset(&sigmask);
	sigaddset(&sigmask, SIGCONT);
	sigaddset(&sigmask, SIGTERM);
	sigaddset(&sigmask, SIGINT);
	
	if (sigwait(&sigmask, &signum) == 0) {
		if (signum == SIGCONT) {
			send_cb(index, errorcount, idlecount);
		} else if ((signum == SIGTERM) || (signum == SIGINT)) {
			return 1;
		}
	}
	
	return 0;
}
