/*
 *      sound.c
 *      
 *      Copyright 2009 - 2011 Alex <alex@linuxonly.ru>
 *      
 *      This program is free software: you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 3 of the License, or
 *      (at your option) any later version.
 *      
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *      
 *      You should have received a copy of the GNU General Public License
 *      along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#include <gmodule.h>
#include <glib.h>
#include <string.h>
#include <stdio.h>
#include <limits.h>
#include <stdint.h>
#include "sound.h"
#include "mainwindow.h"
#include "settings.h"
#include "soundmodules/soundmodules.h"
#include "resources.h"

#define WAV_UNCOMPRESSED_PCM   1

typedef gboolean (*sound_module_init_func)(guint *moduleid, gchar *modulename, gint *moduleversion);
typedef gboolean (*sound_module_play_file_func)(gchar *filename, gint (*wav_header_parser_callback)(gint filefd, struct _wavheader *wavhdr));
typedef gboolean (*sound_module_close_func)(void);

GSList *soundlanguages = NULL;
GSList *soundmodules = NULL;	

sound_module_init_func sound_module_init;
sound_module_play_file_func sound_module_play_file;
sound_module_close_func sound_module_close;

GModule *module;
guint currentmoduleid;
guint currentlanguageid;
gchar *systemlocale;
guchar soundparameter;
guchar soundlimit;


gint sound_wav_header_parser_callback(gint filefd, struct _wavheader *wavhdr)
{
	gchar data[4];
	
	if (wavhdr == NULL) return 0;
	
	if (lseek(filefd, 0, SEEK_CUR) == -1) return 0;
	
	//test if file is wav - RIFF is first 4 bytes
	read(filefd, &data, 4);
	if (strncmp(data, "RIFF", 4) != 0) {
		g_warning("Sound format not supported: no RIFF signature");
		return 0;
	}
	
	lseek(filefd, 4, SEEK_CUR);
	
	//WAVE is third four bytes
	read(filefd, &data, 4);
	if (strncmp(data, "WAVE", 4) != 0) {
		g_warning("Sound format not supported: no WAVE signature");
		return 0;
	}
	
	//find format chunk
	while(1) {
		read(filefd, &data, 4);
		if (strncmp("fmt ", data, 4) == 0) break;
	}
	
	//test format chunk size - must be more than 16 bytes
	read(filefd, &wavhdr->len, 4);
	if (wavhdr->len < 16) {
		g_warning("Sound file corrupted: wrong size of fmt section");
		return 0;
	}
	
	//test format - only uncompressed PCM supported
	read(filefd, &wavhdr->format, 2);
	if (wavhdr->format != WAV_UNCOMPRESSED_PCM) {
		g_warning("Sound format not supported: only uncompressed PCM supported");
		return 0;
	}
	
	//read stream parameters
	read(filefd, &wavhdr->channels, 2);
	//printf("[i] channels = %i\n", (int)wavhdr->channels);
	read(filefd, &wavhdr->samples_per_sec, 4);
	//printf("[i] samples per second = %li\n", wavhdr->samples_per_sec);
	read(filefd, &wavhdr->avg_bytes_per_sec, 4);
	//printf("[i] average bytes per second = %li\n", wavhdr->avg_bytes_per_sec);
	read(filefd, &wavhdr->align, 2);
	read(filefd, &wavhdr->bits, 2);
	//printf("[i] bits per second = %i\n", (int)wavhdr->bits);
	
	//supported bps values - 8 and 16 bps
	if (wavhdr->bits != 8 && wavhdr->bits != 16) {
		g_warning("Sound format not supported: only 8 and 16 bps supported");
		return 0;
	}
	
	//supported channels values - 1 and 2
	if (wavhdr->channels < 1 || wavhdr->channels > 2) {
		g_warning("Sound format not supported: only 1 and 2 channels supported");
		return 0;
	}
	
	//extra bytes at the end of chunk - must be ignored
	wavhdr->len -= 16;
	if (wavhdr->len) lseek(filefd, wavhdr->len, SEEK_CUR);
	
	//find data chunk
	while(1) {
		read(filefd, &data, 4);
		if (strncmp("data", data, 4) == 0) break;
	}
	
	//size of data to read
	read(filefd, &wavhdr->len, 4);
	//printf("[i] data length = %i\n", (int)wavhdr->len);
		
	return 1;
}

guint sound_modules_enum(void)
{
	GDir *moddir;
	const gchar *modname;
	gsize length;
	const gchar modext[] = ".so";
	gchar *modpath;
	guint number = 0;
	struct _soundmodule *soundmodule;
	GModule *module;
	guint moduleid;
	gchar modulename[32];
	gint moduleversion;
	
	
	if ((moddir = g_dir_open(RESOURCE_DIR_SOUND_MODULES, 0, NULL)) != NULL) {
		printf("Sound modules enumeration\n");
		while ((modname = g_dir_read_name(moddir)) != NULL) {
			if (((length = strlen(modname)) > 3) && (strncmp((const char *)modname + length - 3, (const char *)&modext, 3) == 0)) {
				modpath = g_strconcat(RESOURCE_DIR_SOUND_MODULES, modname, NULL);
				if (g_file_test(modpath, G_FILE_TEST_EXISTS)) {
					if ((module = g_module_open(modpath, G_MODULE_BIND_LAZY)) != NULL) {
						if (g_module_symbol(module, "sound_module_init", (gpointer*)&sound_module_init)) {
							if (sound_module_init(&moduleid, (gchar *)&modulename, &moduleversion)) {
								soundmodule = g_malloc0(sizeof(struct _soundmodule));
								soundmodule->path = g_strdup(modpath);
								soundmodule->id = moduleid;
								soundmodule->name = g_strdup(modulename);
								soundmodule->version = moduleversion;
								soundmodules = g_slist_prepend(soundmodules, soundmodule);
								printf(" %u. %s [ID %u]\n", number+1, modulename, moduleid);
								number++;
							}
						}
						if (!g_module_close(module)) g_warning("Error unloading sound module: %s\n", (gchar*)g_module_error()); 
					}
				}
				g_free(modpath);
			}
		}
		g_dir_close(moddir);
	} else {
		g_warning("No sound modules found");
	}
	
	if (number > 0) soundmodules = g_slist_reverse(soundmodules);
	
	return number;
} 

void sound_modules_free_foreach(gpointer data, gpointer user_data)
{
	struct _soundmodule *soundmodule;
	
	soundmodule = (struct _soundmodule *)data;
	
	if (soundmodule->path != NULL) g_free(soundmodule->path);
	if (soundmodule->name != NULL) g_free(soundmodule->name);
	
}

gboolean sound_modules_free(GSList *modules)
{
	if (modules == NULL) return FALSE;
	
	g_slist_foreach(modules, sound_modules_free_foreach, NULL);
	
	g_slist_free(modules);
	
	return TRUE;
}


gboolean sound_modules_set_module(guint moduleid)
{
	gint i;
	struct _soundmodule *soundmodule;
	gsize length;
	gboolean modulefound = FALSE;
		
	if (currentmoduleid == moduleid) return FALSE;
	
	if ((length = g_slist_length(soundmodules)) > 0) {
		for (i=0; i<length; i++) {
			soundmodule = (struct _soundmodule *)g_slist_nth_data(soundmodules, i);
			if (soundmodule->id == moduleid) {
				//clean module unload
				if (module != NULL) {
					sound_module_close();
					if (!g_module_close(module)) g_warning("Error unloading sound module: %s\n", (gchar*)g_module_error());
				} 
				//new module
				if ((module = g_module_open(soundmodule->path, G_MODULE_BIND_LAZY)) != NULL) {
					if (g_module_symbol(module, "sound_module_play_file", (gpointer*)&sound_module_play_file)) {
						g_module_symbol(module, "sound_module_close", (gpointer*)&sound_module_close);
						currentmoduleid = moduleid;
						modulefound = TRUE;
					}
				}
				break;
			}
		}
	} else return FALSE;
	
	if (!modulefound) {
		soundmodule = (struct _soundmodule *)g_slist_nth_data(soundmodules, 0);
		//clean module unload
		if (module != NULL) {
			sound_module_close();
			if (!g_module_close(module)) g_warning("Error unloading sound module: %s\n", (gchar*)g_module_error());
		} 
		//new module
		if ((module = g_module_open(soundmodule->path, G_MODULE_BIND_LAZY)) != NULL) {
			if (g_module_symbol(module, "sound_module_play_file", (gpointer*)&sound_module_play_file)) {
				g_module_symbol(module, "sound_module_close", (gpointer*)&sound_module_close);
				currentmoduleid = soundmodule->id;
				modulefound = TRUE;
			}
		}
	}
	
	if (modulefound) {
		settings_append_int("sounds_module_id", currentmoduleid);
	}
	
	return modulefound;
}

void sound_langauages_free_foreach(gpointer data, gpointer user_data)
{
	struct _soundlanguage *soundlanguage;
	
	soundlanguage = (struct _soundlanguage *)data;
	g_free(soundlanguage->code);
}

gboolean sound_languages_enum(void)
{
	GDir *langdir;
	const gchar *langname;
	struct _soundlanguage *soundlanguage;
	gchar **localev;
	gboolean aslocale = FALSE;
	const gchar *enlanguage = "en_US";
	gchar *setlanguage;
	gchar path[PATH_MAX];
	gchar filepath[PATH_MAX];
	gint langid = 0;
	gint langlocaleid = -1;
	gint langsetid =-1;
	gint langenid =-1;

	g_slist_foreach(soundlanguages, sound_langauages_free_foreach, NULL);
	g_slist_free(soundlanguages); soundlanguages = NULL;
	if (systemlocale != NULL) {
		g_free(systemlocale);
		systemlocale = NULL;
	}
	
	localev = g_strsplit(g_getenv("LANG"), ".", 2);
	if (localev[0] != NULL) systemlocale = g_strdup(localev[0]); else systemlocale = g_strdup(enlanguage);
	g_strfreev(localev);
	printf("System locale: %s\n", systemlocale); 
	
	setlanguage = settings_read_string("sounds_language", "locale");
	if (g_str_equal(setlanguage, "locale")) {
		g_free(setlanguage);
		setlanguage = g_strdup(systemlocale);
		aslocale = TRUE;
	}
	
	if ((langdir = g_dir_open(RESOURCE_DIR_SOUND_LANGUAGES, 0, NULL)) == NULL) {
		g_warning("No languages found");
		return FALSE;
	} else {
		printf("Languages sound packs enumeration\n");
		while ((langname = g_dir_read_name(langdir)) != NULL) {
			g_sprintf(path, "%s%s", RESOURCE_DIR_SOUND_LANGUAGES, langname);
			if (g_file_test(path, G_FILE_TEST_EXISTS)) {
				soundlanguage = (struct _soundlanguage *)g_malloc0(sizeof(struct _soundlanguage));
				soundlanguage->code = g_strdup(langname);
				soundlanguage->multiplex = SOUND_LANGUAGE_MULTIPLEX_VOID;
				
				g_sprintf(filepath, "%s/10.wav", path);
				if (g_file_test(filepath, G_FILE_TEST_EXISTS))
					soundlanguage->multiplex = SOUND_LANGUAGE_MULTIPLEX_10;
				g_sprintf(filepath, "%s/100.wav", path);
				if (g_file_test(filepath, G_FILE_TEST_EXISTS)) 
					soundlanguage->multiplex = SOUND_LANGUAGE_MULTIPLEX_100;
								
				if (soundlanguage->multiplex == SOUND_LANGUAGE_MULTIPLEX_VOID) {
					g_free(soundlanguage);
				} else {
					soundlanguages = g_slist_append(soundlanguages, soundlanguage);
					if ((langlocaleid == -1) && (g_str_equal(systemlocale, langname)))
						langlocaleid = langid;
					if ((langsetid == -1) && (g_str_equal(setlanguage, langname)))
						langsetid = langid;
					if ((langenid == -1) && (g_str_equal(enlanguage, langname)))
						langenid = langid;
					printf(" %u. %s\n", langid+1, langname);
					langid++;
				}
			}
		}
		g_dir_close(langdir);
	}
	
	if (g_slist_length(soundlanguages) == 0) {
		g_warning("No languages found");
		return FALSE;
	}
	
	if ((aslocale) && (langlocaleid != -1)) {
		currentlanguageid = langlocaleid;
		return TRUE;
	}
	
	if (langsetid != -1) {
		currentlanguageid = langsetid;
		return TRUE;
	} else if (langenid != -1) {
		currentlanguageid = langenid;
		return TRUE;
	} else {
		currentlanguageid = 0;
		return TRUE;
	}
	
	return FALSE;
}

gboolean sound_languages_set_language(guint languageid)
{
	gsize length;
	gint i;
	struct _soundlanguage *soundlanguage;
	gboolean languagefound = FALSE;
	
	length = g_slist_length(soundlanguages);
	
	if ((soundlanguages == NULL) || (languageid > length)) return FALSE;
	
	if (languageid == length) {
		for (i=0; i<length; i++) {
			soundlanguage = (struct _soundlanguage *)g_slist_nth_data(soundlanguages, i);
			if (g_str_equal(systemlocale, soundlanguage->code)) {
				currentlanguageid = i;
				languagefound = TRUE;
				break;
			}
		}
		if (!languagefound) currentlanguageid = 0;
		settings_append_string("sounds_language", "locale");
	} else {
		currentlanguageid = languageid;
		soundlanguage = (struct _soundlanguage *)g_slist_nth_data(soundlanguages, languageid);
		settings_append_string("sounds_language", soundlanguage->code);
	}
	
	return TRUE;
}

gchar **sound_languages_path(struct _frontend *frontend, guchar parameter, guchar limit)
{
	guint reallevel;
	guint multiplex;
	guint limitlevel;
	gchar **pathv;
	struct _soundlanguage *soundlanguage;
	
	if (soundlanguages != NULL) {
		soundlanguage = (struct _soundlanguage *)g_slist_nth_data(soundlanguages, currentlanguageid);
		if (soundlanguage->multiplex == SOUND_LANGUAGE_MULTIPLEX_100) {
			multiplex = 100;
		} else if (soundlanguage->multiplex == SOUND_LANGUAGE_MULTIPLEX_10) {
			multiplex = 10;
		}
	} else {
		return NULL;
	}
	
	if (frontend->tuned) {
		switch (parameter) {
			case SOUND_PARAMETER_SNR: {
				reallevel = (frontend->snr/settings_read_float("scale_snr", 100.0)) * multiplex;
				break;
			}
			
			case SOUND_PARAMETER_AGC: {
				reallevel = (frontend->signal/settings_read_float("scale_agc", 1000.0)) * multiplex;
				break;
			}
		
			case SOUND_PARAMETER_BER: {
				reallevel = (frontend->ber/settings_read_float("scale_ber", 4294967296.0)) * multiplex;
				break;
			}
			
			default: {
				reallevel = 0;
				break;
			}
		}
		if (reallevel > multiplex) reallevel = multiplex;
	} else {
		reallevel = 0;
	}
	
	if (soundlanguage->multiplex == SOUND_LANGUAGE_MULTIPLEX_10) {
		if ((reallevel * 10) < limit) return NULL;
	} else if (soundlanguage->multiplex == SOUND_LANGUAGE_MULTIPLEX_100) {
		if (reallevel < limit) return NULL;
	}
	
	if (soundlanguage->multiplex == SOUND_LANGUAGE_MULTIPLEX_10) {
		pathv = g_malloc(sizeof(gchar *)*2);
		pathv[0] = g_strdup_printf("%s%s/%u.wav", RESOURCE_DIR_SOUND_LANGUAGES, soundlanguage->code, reallevel);
		pathv[1] = NULL;
	} else if (soundlanguage->multiplex == SOUND_LANGUAGE_MULTIPLEX_100) {
		if ((reallevel <= 20) || (reallevel % 10 == 0)) {
			pathv = g_malloc(sizeof(gchar *)*2);
			pathv[0] = g_strdup_printf("%s%s/%u.wav", RESOURCE_DIR_SOUND_LANGUAGES, soundlanguage->code, reallevel);
			pathv[1] = NULL;
		} else {
			pathv = g_malloc(sizeof(gchar *)*3);
			pathv[0] = g_strdup_printf("%s%s/%u.wav", RESOURCE_DIR_SOUND_LANGUAGES, soundlanguage->code, reallevel - reallevel % 10);
			pathv[1] = g_strdup_printf("%s%s/%u.wav", RESOURCE_DIR_SOUND_LANGUAGES, soundlanguage->code, reallevel % 10);
			pathv[2] = NULL;
		}
	}
	
	if (!g_file_test(pathv[0], G_FILE_TEST_EXISTS)) {
		g_strfreev(pathv);
		return NULL;
	}
	
	return pathv;
}

void sound_limit_set(guchar limit)
{
	if (limit <= 100) {
		soundlimit = limit;
		settings_append_int("sounds_limit", soundlimit);
	}
}

guint sound_limit_get(void)
{
	return soundlimit;
}

void sound_parameter_set(guchar parameter)
{
	switch (parameter) {
		case 0: soundparameter = SOUND_PARAMETER_SNR; break;
		case 1: soundparameter = SOUND_PARAMETER_AGC; break;
		case 2: soundparameter = SOUND_PARAMETER_BER; break;
		default: soundparameter = SOUND_PARAMETER_SNR; break;
	}
	
	settings_append_int("sounds_parameter", (guint)soundparameter);
}

guint sound_parameter_get(void)
{
	return soundparameter;
}

GSList *sound_modules_get(guint *index)
{
	gint i;
	struct _soundmodule *soundmodule;
	gsize length;
		
	if (index != NULL) {
		if ((length = g_slist_length(soundmodules)) > 0) {
			for (i=0; i<length; i++) {
				soundmodule = (struct _soundmodule *)g_slist_nth_data(soundmodules, i);
				if (soundmodule->id == currentmoduleid) {
					*index = i;
					break;
				}
			}
		}
	}
		
	return soundmodules;
}

GSList *sound_languages_get(guint *index)
{
	gchar *setlang;
	
	if (index != NULL) {
		if (soundlanguages != NULL) {
			setlang = settings_read_string("sounds_language", "locale");
			if (g_str_equal(setlang, "locale")) {
				*index = g_slist_length(soundlanguages);
			} else {
				*index = currentlanguageid;
			}
			g_free(setlang);
		} else {
			*index = 0;
		}
	}
	
	return soundlanguages;
}

gboolean sound_init(void)
{
	gboolean langres;
	guint modules;
	guint setmoduleid;
	gboolean modres = FALSE;
	
	langres = sound_languages_enum();
	modules = sound_modules_enum();
		
	setmoduleid = settings_read_int("sounds_module_id", 1);
	
	if (modules > 0) {
		if (sound_modules_set_module(setmoduleid)) {
			modres = TRUE;
		}
	}
	
	soundparameter = settings_read_int("sounds_parameter", 0);
	soundlimit = settings_read_int("sounds_limit", 0);
	
	return langres && modres;
}

gboolean sound_play(struct _frontend *frontend, gulong timeout)
{
	gchar **sndpathv;
	guint i;
	
	if ((sndpathv = sound_languages_path(frontend, soundparameter, soundlimit)) != NULL) {
		i = 0;
		while (sndpathv[i] != NULL) {
			//printf("--> %s\n", sndpathv[i]);
			sound_module_play_file(sndpathv[i], sound_wav_header_parser_callback);
			i++;
		}
		g_strfreev(sndpathv);
		if (timeout != 0) g_usleep(timeout);
		return TRUE;
	}
	
	if (timeout != 0) g_usleep(timeout);
	
	return FALSE;
}
