/*
 *      streamconfig.c
 *      
 *      Copyright 2009 - 2011 Alex <alex@linuxonly.ru>
 *      
 *      This program is free software: you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 3 of the License, or
 *      (at your option) any later version.
 *      
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *      
 *      You should have received a copy of the GNU General Public License
 *      along with this program. If not, see <http://www.gnu.org/licenses/>.
 */


#include <glib.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <string.h>
#include <errno.h>
#include <netinet/in.h>
#include <sys/ioctl.h>
#include <net/if.h>
#include <net/route.h>
#include <arpa/inet.h>
#include <stdint.h>
#include <sys/un.h>
#include <sys/wait.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include "streamconfig.h"
#include "mainwindow.h"
#include "resources.h"
#include "./common/serialize.h"
#include "./gsfprivileged/gsfshm.h"

#define SHORT_ANSWER_BUFFER   16
#define ANSWER_BUFFER         1024

static GSList *netinterfaces = NULL;

static gboolean streamconfig_request(struct _streamconfig *sc, char *reqbuf, size_t *reqsize, char *ansbuf, size_t *anssize);
gboolean streamconfig_file_flush(struct _streamconfig *sc);
gboolean streamconfig_file_append_string(struct _streamconfig *sc, gchar *name, gchar *value);
gchar *streamconfig_file_read_string(struct _streamconfig *sc, gchar *name, gchar *def);
gboolean streamconfig_file_append_int(struct _streamconfig *sc, gchar *name, gint value);
gint streamconfig_file_read_int(struct _streamconfig *sc, gchar *name, gint def);


gboolean streamconfig_get_running(struct _streamconfig *sc)
{
	gint lockfd;
	const gchar *homedir;
	gchar *lockfn;
	struct flock lock;
	
	if ((homedir = g_getenv("HOME")) == NULL) homedir = g_get_home_dir();
	lockfn = g_strdup_printf(STREAMCONFIG_LOCK_PATH, homedir);
	
	if ((lockfd = open(lockfn, O_RDWR, 0)) == -1) {
		g_free(lockfn);
		return FALSE;
	}
	g_free(lockfn);
	
	lock.l_type = F_WRLCK;
	lock.l_whence = SEEK_SET;
	lock.l_start = 0;
	lock.l_len = 0;
	if (fcntl(lockfd, F_SETLK, &lock) == -1) {
		if (errno == EACCES || errno == EAGAIN) {
			if (fcntl(lockfd, F_GETLK, &lock) != -1) {
				if (sc != NULL) {
					sc->dpid = lock.l_pid;
				}
				close(lockfd);
				return TRUE;
			}
			close(lockfd);
			return TRUE;
		}
	}
	
	close(lockfd);
	return FALSE;
}

gchar *streamconfig_sockaddr_to_string(struct sockaddr_in *sa, gchar *buffer, gsize length)
{
	switch (sa->sin_family) {
			case AF_INET: {
				inet_ntop(AF_INET, &sa->sin_addr, buffer, length);
				break;
			}
			
			case AF_INET6: {
				inet_ntop(AF_INET6, &sa->sin_addr, buffer, length);
				break;
			}
			
			default: {
				inet_ntop(AF_INET, &sa->sin_addr, buffer, length);
				break;
			}
		}
	
	return buffer;
}

void streamconfig_interfaces_get_with_multicast_foreach(gpointer data, GSList **interfaces)
{
	struct _netinterface *interface;
	
	interface = (struct _netinterface *)data;
	if ((interface->ifup) && ((interface->ifmc) || (interface->ifloop))) {
		*interfaces = g_slist_prepend(*interfaces, interface);
	}
}

GSList *streamconfig_interfaces_get_with_multicast(void)
{
	GSList *interfaces = NULL;
	
	if (netinterfaces == NULL) return NULL;
	
	g_slist_foreach(netinterfaces, (GFunc)streamconfig_interfaces_get_with_multicast_foreach, &interfaces);
	interfaces = g_slist_reverse(interfaces);
	
	return interfaces;
}

void streamconfig_interfaces_get_with_broadcast_foreach(gpointer data, GSList **interfaces)
{
	struct _netinterface *interface;
	
	interface = (struct _netinterface *)data;
	if ((interface->ifup) && ((interface->ifbc) || (interface->ifloop))) {
		*interfaces = g_slist_prepend(*interfaces, interface);
	}
}

GSList *streamconfig_interfaces_get_with_broadcast(void)
{
	GSList *interfaces = NULL;
	
	if (netinterfaces == NULL) return NULL;
	
	g_slist_foreach(netinterfaces, (GFunc)streamconfig_interfaces_get_with_broadcast_foreach, &interfaces);
	interfaces = g_slist_reverse(interfaces);
	
	return interfaces;
}

void streamconfig_interfaces_list_destroy_foreach(gpointer data, gpointer user_data)
{
	g_slice_free(struct _netinterface, data);
}

void streamconfig_interfaces_list_destroy(void)
{
	g_slist_foreach(netinterfaces, streamconfig_interfaces_list_destroy_foreach, NULL);
	g_slist_free(netinterfaces);
	netinterfaces = NULL;
}

gint streamconfig_interfaces_list_build(void)
{
	gint socketfd;
	struct ifconf ifc;
	struct ifreq  *ifr, *currentifr;
	gint ifcbuflen;
	gchar *ifcbuf;
	guint ifcount;
	gint i;
	struct _netinterface *interface;
	
	
	if ((socketfd = socket(AF_INET, SOCK_DGRAM, 0)) == -1) {
		perror("Socket creation error");
		return -1;
	}
	
	if (netinterfaces != NULL)
		streamconfig_interfaces_list_destroy();
	
	ifcbuflen = 100 * sizeof(struct ifreq);
	
	while (1) {
		ifcbuf = malloc(ifcbuflen);
		ifc.ifc_len = ifcbuflen;
		ifc.ifc_buf = ifcbuf;
		
		if (ioctl(socketfd, SIOCGIFCONF, &ifc) == -1) {
			if (errno != EINVAL) perror("IOCTL SIOCGIFCONF error");
		} else {
			ifcbuflen = ifc.ifc_len;
			break;
		}		 
		ifcbuflen += 10 * sizeof(struct ifreq);
		free(ifcbuf);
	}
	
	ifr = ifc.ifc_req;
	ifcount = ifc.ifc_len / sizeof(struct ifreq);
	
	for (i=0; i<ifcount; i++) {
		currentifr = &ifr[i];
		
		interface = g_slice_new0(struct _netinterface);
				
		memcpy(interface->ifname, currentifr->ifr_name, IFNAMSIZ);
		
		streamconfig_sockaddr_to_string((struct sockaddr_in *)&currentifr->ifr_addr, (char *)&interface->ifipaddr, INET6_ADDRSTRLEN);
		
		interface->ifindex = if_nametoindex((gchar *)&interface->ifname);
		
		if (ioctl(socketfd, SIOCGIFHWADDR, currentifr) == -1) {
			perror("IOCTL SIOCGIFHWADDR error");
		} else {
			memcpy(interface->ifmacaddr, &currentifr->ifr_hwaddr.sa_data[0], IFHWADDRLEN);
		}
		
		if (ioctl(socketfd, SIOCGIFFLAGS, currentifr) == -1) {
			interface->ifup = interface->ifbc = interface->ifmc = interface->ifloop = interface->ifptp = 0;
			perror("IOCTL SIOCGIFFLAGS error");
		} else {
			if (currentifr->ifr_flags & IFF_UP) interface->ifup = 1; else interface->ifup = 0;
			if (currentifr->ifr_flags & IFF_BROADCAST) interface->ifbc = 1; else interface->ifbc = 0;
			if (currentifr->ifr_flags & IFF_MULTICAST) interface->ifmc = 1; else interface->ifmc = 0;
			if (currentifr->ifr_flags & IFF_LOOPBACK) interface->ifloop = 1; else interface->ifloop = 0;
			if (currentifr->ifr_flags & IFF_POINTOPOINT) interface->ifptp = 1; else interface->ifptp = 0;
		}
		
		if (ioctl(socketfd, SIOCGIFBRDADDR, currentifr) == -1) {
			perror("IOCTL SIOCGIFBRDADDR error");
		} else {
			streamconfig_sockaddr_to_string((struct sockaddr_in *)&currentifr->ifr_addr, (char *)&interface->ifbcipaddr, INET_ADDRSTRLEN);
		}
		
		if (ioctl(socketfd, SIOCGIFNETMASK, currentifr) == -1) {
			perror("IOCTL SIOCGIFBRDADDR error");
		} else {
			streamconfig_sockaddr_to_string((struct sockaddr_in *)&currentifr->ifr_netmask, (char *)&interface->ifnetmask, INET6_ADDRSTRLEN);
		}
		
		netinterfaces = g_slist_prepend(netinterfaces, interface);
		
		printf("%s <%u>:\n", interface->ifname, interface->ifindex);
		printf("  IP: %s, NETMASK: %s, BC IP: %s, MAC: %.2x:%.2x:%.2x:%.2x:%.2x:%.2x\n", interface->ifipaddr, interface->ifnetmask, interface->ifbcipaddr, interface->ifmacaddr[0], interface->ifmacaddr[1], interface->ifmacaddr[2], interface->ifmacaddr[3], interface->ifmacaddr[4], interface->ifmacaddr[5]);
		printf("  [UP: %i, BC: %i, MC: %i, LOOP: %i, PTP: %i]\n", interface->ifup, interface->ifbc, interface->ifmc, interface->ifloop, interface->ifptp);
	}
	
	free(ifcbuf);
	close(socketfd);
	
	netinterfaces = g_slist_reverse(netinterfaces);
		
	return ifcount;
}


gboolean streamconfig_add_multicast_route(struct _streamconfig *sc, gchar *intname)
{
	key_t shmkey;
	int shmid;
	struct srequest *shmblock;
	gchar **ppargv;
	gint ppstatus;
	GError *pperror = NULL;
	gboolean result = FALSE; 
	
	
	if (strncmp(sc->mcintname, intname, IFNAMSIZ) == 0) return TRUE;
	
	shmkey = ftok(PRIVILEGED_TOKEN_FILE, PRIVILEGED_PROJID);
	shmid = shmget(shmkey, sizeof(struct srequest), 0666 | IPC_CREAT);
	shmblock = (struct srequest *)shmat(shmid, 0, 0);
	
	memset(shmblock, 0, sizeof(struct srequest));
	shmblock->action = PRIVILEGED_ACTION_ADD_MULTICAST_ROUTE;
	strncpy(shmblock->intname, intname, IFNAMSIZ);
	strncpy(shmblock->remintname, sc->mcintname, IFNAMSIZ);
	
	ppargv = (gchar **)g_malloc(sizeof(gchar *));
	
	ppargv[0] = g_strdup(RESOURCE_BINARY_PRIVILEGED); 
	ppargv[1] = NULL;
	if (g_spawn_sync(NULL, ppargv, NULL, G_SPAWN_STDOUT_TO_DEV_NULL | G_SPAWN_STDERR_TO_DEV_NULL | G_SPAWN_FILE_AND_ARGV_ZERO, NULL, NULL, NULL, NULL, &ppstatus, &pperror)) {
		ppstatus = WEXITSTATUS(ppstatus);
		switch(ppstatus) {
			case PRIVILEGED_STATUS_OK:
				strncpy(sc->mcintname, intname, IFNAMSIZ);
				printf("[+] Set multicast route for interface %s\n", intname);
				break;
			case PRIVILEGED_STATUS_ROUTE_ERROR:
				printf("[-] Failed to set multicast route for interface %s\n", intname);
				break;
			default: 
				printf("[-] Failed to set multicast route for interface %s (status %i)\n", intname, ppstatus);
				break;
		}
	} else {
		printf("[-] Error while executing privileged process: %s\n", pperror->message);
		g_error_free(pperror);
	}
	
	g_strfreev(ppargv);
	
	shmdt(shmblock);
	shmctl(shmid, IPC_RMID, 0);
		
	return result;
}

gboolean streamconfig_remove_multicast_route(struct _streamconfig *sc)
{
	key_t shmkey;
	int shmid;
	struct srequest *shmblock;
	gchar **ppargv;
	gint ppstatus;
	GError *pperror = NULL;
	gboolean result = FALSE; 
	
	
	if (sc->mcintname[0] == '\0') return TRUE;
	
	shmkey = ftok(PRIVILEGED_TOKEN_FILE, PRIVILEGED_PROJID);
	shmid = shmget(shmkey, sizeof(struct srequest), 0666 | IPC_CREAT);
	shmblock = (struct srequest *)shmat(shmid, 0, 0);
	
	memset(shmblock, 0, sizeof(struct srequest));
	shmblock->action = PRIVILEGED_ACTION_REMOVE_MULTICAST_ROUTE;
	strncpy(shmblock->intname, sc->mcintname, IFNAMSIZ);
	
	ppargv = (gchar **)g_malloc(sizeof(gchar *));
	
	ppargv[0] = g_strdup(RESOURCE_BINARY_PRIVILEGED); 
	ppargv[1] = NULL;
	if (g_spawn_sync(NULL, ppargv, NULL, G_SPAWN_STDOUT_TO_DEV_NULL | G_SPAWN_STDERR_TO_DEV_NULL | G_SPAWN_FILE_AND_ARGV_ZERO, NULL, NULL, NULL, NULL, &ppstatus, &pperror)) {
		ppstatus = WEXITSTATUS(ppstatus);
		switch(ppstatus) {
			case PRIVILEGED_STATUS_OK:
				printf("[+] Removed multicast route for interface %s\n", sc->mcintname);
				memset(sc->mcintname, 0, IFNAMSIZ);
				break;
			case PRIVILEGED_STATUS_ROUTE_ERROR:
				printf("[-] Failed to remove multicast route for interface %s\n", sc->mcintname);
				break;
			default: 
				printf("[-] Failed to remove multicast route for interface %s (status %i)\n", sc->mcintname, ppstatus);
				break;
		}
	} else {
		printf("[-] Error while executing privileged process: %s\n", pperror->message);
		g_error_free(pperror);
	}
	
	g_strfreev(ppargv);
	
	shmdt(shmblock);
	shmctl(shmid, IPC_RMID, 0);
		
	return result;
}

gboolean streamconfig_init(struct _streamconfig *sc)
{
	const gchar *fullusername;
	gchar username[82];
	gsize unamelength;
	gint fd;
	GError *fileerror = NULL;
	GMappedFile *mmap;
	gchar **processargv;
	GError *process_error = NULL;
	gint status;
	gchar *str;
	guint length;
	gchar mcintname[IFNAMSIZ];
	
	sc->serverctl = FALSE;
	
	if ((fullusername = getenv("USER")) == NULL) fullusername = g_get_user_name();
	
	if ((unamelength = strlen(fullusername)) >= 82) {
		strncpy(username, fullusername, 81);
		username[81] = '\0';
	} else {
		strncpy(username, fullusername, unamelength);
		username[unamelength] = '\0';
	}
	
	sc->clientsockpath = g_strdup_printf(STREAMCONFIG_CLIENT_SOCKET, username);
	sc->serversockpath = g_strdup_printf(STREAMCONFIG_SERVER_SOCKET, username);
	
	memset(&sc->mcintname, 0, sizeof(sc->mcintname));
	str = streamconfig_file_read_string(sc, "udp_multicast_interface", "lo");
	if ((length = strlen(str)) > IFNAMSIZ) {
		strncpy((gchar *)&mcintname, str, IFNAMSIZ);
		mcintname[IFNAMSIZ] = '\0';
	} else {
		strncpy((gchar *)&mcintname, str, length);
		mcintname[length] = '\0';
	}
		
	if ((sc->clientsock = socket(AF_LOCAL, SOCK_DGRAM, 0)) == -1) {
		perror("[-] Creating stream server control socket");
		return FALSE;
	}
	
	unlink(sc->clientsockpath);
	
	memset(&sc->clientaddr, 0, sizeof(struct sockaddr_un));
	sc->clientaddr.sun_family = AF_LOCAL;
	strncpy(sc->clientaddr.sun_path, sc->clientsockpath, strlen(sc->clientsockpath));
	
	if (bind(sc->clientsock, (struct sockaddr *)&sc->clientaddr, sizeof(struct sockaddr_un)) == -1) {
		perror("[-] Binding stream server control socket");
		close(sc->clientsock);
		return FALSE;
	}
	
	memset(&sc->serveraddr, 0, sizeof(struct sockaddr_un));
	sc->serveraddr.sun_family = AF_LOCAL;
	strncpy(sc->serveraddr.sun_path, sc->serversockpath, strlen(sc->serversockpath));
	
	if ((g_mkdir_with_parents(g_build_filename(g_get_home_dir(), STREAMCONFIG_CONFDIR_PATH, NULL), STREAMCONFIG_DIR_PERMISSIONS)) == -1) {
		sc->fileopened = FALSE;
		return FALSE;
	}
	
	sc->filename = g_build_filename(g_get_home_dir(), STREAMCONFIG_CONFFILE_PATH, NULL);
	sc->configfile = g_key_file_new();
	
	if (!g_file_test(sc->filename, G_FILE_TEST_EXISTS)) {
		if ((fd = g_creat(sc->filename, S_IRUSR | S_IWUSR)) != -1) {
			close(fd);
		} else {
			sc->fileopened = FALSE;
			return FALSE;
		}
	}
		
	if ((mmap = g_mapped_file_new(sc->filename, FALSE, &fileerror)) != NULL) {
		if (g_mapped_file_get_length(mmap) > 0) {
			g_key_file_load_from_data(sc->configfile, g_mapped_file_get_contents(mmap), g_mapped_file_get_length(mmap), G_KEY_FILE_NONE, &fileerror);
		}
		g_mapped_file_free(mmap);
	}
	
	sc->fileopened = TRUE;
	
	
	streamconfig_interfaces_list_build();
	
	
	if (!streamconfig_get_running(sc)) {
		if (g_file_test(RESOURCE_BINARY_STREAMSERVER, G_FILE_TEST_IS_EXECUTABLE)) {
			streamconfig_add_multicast_route(sc, (gchar *)&mcintname);
			
			processargv = (gchar **)g_malloc(sizeof(gchar *)*2);
						
			processargv[0] = g_strdup(RESOURCE_BINARY_STREAMSERVER);
			processargv[1] = g_strdup("-d");
			processargv[2] = NULL;
					
			if (!g_spawn_sync(NULL, processargv, NULL, 0, NULL, NULL, NULL, NULL, &status, &process_error)) {
				printf("[-] Error while running streaming server: %s\n", process_error->message);
				g_error_free(process_error);
				g_strfreev(processargv);
				return FALSE;
			}
			g_strfreev(processargv);
			status = WEXITSTATUS(status);
						
			printf("[+] Started streaming server\n");
			sc->serverctl = TRUE;
		}
	}
		
	return TRUE;
}

gboolean streamconfig_close(struct _streamconfig *sc)
{
	GError *fileerror = NULL;
	gchar *filecontent;
	gsize filebytes;
	gboolean result = FALSE;
	
	if (sc->clientsock != -1) close(sc->clientsock);
	
	unlink(sc->clientsockpath);
	
	if (sc->clientsockpath != NULL) g_free(sc->clientsockpath);
	if (sc->serversockpath != NULL) g_free(sc->serversockpath);
	
	if (sc->fileopened) {
		filecontent = g_key_file_to_data(sc->configfile, &filebytes, NULL);
		result = g_file_set_contents(sc->filename, filecontent, -1, &fileerror);
		g_free(sc->filename);
		g_free(filecontent);
		g_key_file_free(sc->configfile);
		sc->fileopened = FALSE;
	}
	
	if ((streamconfig_get_running(sc)) && (sc->serverctl)) {
		streamconfig_remove_multicast_route(sc);
		kill(sc->dpid, SIGTERM);
	}
	
	return TRUE;
}

gboolean streamconfig_file_flush(struct _streamconfig *sc)
{
	GError *fileerror = NULL;
	gboolean result = FALSE;
	gchar *filecontent;
	gsize filebytes;
	
	if (sc->fileopened) {
		filecontent = g_key_file_to_data(sc->configfile, &filebytes, NULL);
		result = g_file_set_contents(sc->filename, filecontent, filebytes, &fileerror);
	}
	
	return result;
} 

gboolean streamconfig_file_append_string(struct _streamconfig *sc, gchar *name, gchar *value)
{
	gboolean result = FALSE;
	
	if (sc->fileopened) {
		g_key_file_set_string(sc->configfile, STREAMCONFIG_FILE_SECTION, name, value);
		result = TRUE;
	}
	
	return result;
}

gchar *streamconfig_file_read_string(struct _streamconfig *sc, gchar *name, gchar *def)
{
	GError *fileerror = NULL;
	gchar *result = def;
	
	if (sc->fileopened) {
		if (g_key_file_has_key(sc->configfile, STREAMCONFIG_FILE_SECTION, name, &fileerror)) {
			if (fileerror != NULL) fileerror = NULL;
			result = g_key_file_get_string(sc->configfile, STREAMCONFIG_FILE_SECTION, name, &fileerror);
		}
	}
	
	return result;
}

gboolean streamconfig_file_append_int(struct _streamconfig *sc, gchar *name, gint value)
{
	gboolean result = FALSE;
	
	if (sc->fileopened) {
		g_key_file_set_integer(sc->configfile, STREAMCONFIG_FILE_SECTION, name, value);
		result = TRUE;
	}
	
	return result;
}

gint streamconfig_file_read_int(struct _streamconfig *sc, gchar *name, gint def)
{
	GError *fileerror = NULL;
	gint result = def;
	
	if (sc->fileopened) {
		if (g_key_file_has_key(sc->configfile, STREAMCONFIG_FILE_SECTION, name, &fileerror)) {
			if (fileerror != NULL) fileerror = NULL;
			result = g_key_file_get_integer(sc->configfile, STREAMCONFIG_FILE_SECTION, name, &fileerror);
		}
	}
	
	return result;
}

gboolean streamconfig_set_record_status(struct _streamconfig *sc, gboolean status)
{
	const char set_record_status_req[] = SET_RECORD_STATUS_REQ;
	char reqid = STREAMCONFIG_SET_RECORD_STATUS;
	char reqstatus = (char)status;
	char *reqbuf;
	size_t reqsize;
	char ansbuf[SHORT_ANSWER_BUFFER];
	size_t anssize = SHORT_ANSWER_BUFFER;
	
	if ((reqbuf = serialize_form_message((char *)&set_record_status_req, 0, &reqsize, reqid, reqstatus)) != NULL) {
		if ((streamconfig_request(sc, reqbuf, &reqsize, (char *)&ansbuf, &anssize)) && (ansbuf[1] == STREAMCONFIG_RES_SUCCESS)) {
			streamconfig_file_append_int(sc, "record_status", (gint)status);
			streamconfig_file_flush(sc);
			return TRUE;
		}
	}
	
	return FALSE;
}

gboolean streamconfig_set_udp_broadcast_status(struct _streamconfig *sc, gboolean status)
{
	const char set_udpbc_status_req[] = SET_UDPBC_STATUS_REQ;
	char reqid = STREAMCONFIG_SET_UDP_BC_STATUS;
	char reqstatus = (char)status;
	char *reqbuf;
	size_t reqsize;
	char ansbuf[SHORT_ANSWER_BUFFER];
	size_t anssize = SHORT_ANSWER_BUFFER;
	
	if ((reqbuf = serialize_form_message((char *)&set_udpbc_status_req, 0, &reqsize, reqid, reqstatus)) != NULL) {
		if ((streamconfig_request(sc, reqbuf, &reqsize, (char *)&ansbuf, &anssize)) && (ansbuf[1] == STREAMCONFIG_RES_SUCCESS)) {
			streamconfig_file_append_int(sc, "udp_broadcast_status", (gint)status);
			streamconfig_file_flush(sc);
			return TRUE;
		}
	}
	
	return FALSE;
}

gboolean streamconfig_set_udp_multicast_status(struct _streamconfig *sc, gboolean status)
{
	const char set_udpmc_status_req[] = SET_UDPMC_STATUS_REQ;
	char reqid = STREAMCONFIG_SET_UDP_MC_STATUS;
	char reqstatus = (char)status;
	char *reqbuf;
	size_t reqsize;
	char ansbuf[SHORT_ANSWER_BUFFER];
	size_t anssize = SHORT_ANSWER_BUFFER;
	
	if ((reqbuf = serialize_form_message((char *)&set_udpmc_status_req, 0, &reqsize, reqid, reqstatus)) != NULL) {
		if ((streamconfig_request(sc, reqbuf, &reqsize, (char *)&ansbuf, &anssize)) && (ansbuf[1] == STREAMCONFIG_RES_SUCCESS)) {
			streamconfig_file_append_int(sc, "udp_multicast_status", (gint)status);
			streamconfig_file_flush(sc);
			return TRUE;
		}
	}
	
	return FALSE;
}

gboolean streamconfig_set_tcp_unicast_status(struct _streamconfig *sc, gboolean status)
{
	const char set_tcpuc_status_req[] = SET_TCPUC_STATUS_REQ;
	char reqid = STREAMCONFIG_SET_TCP_UC_STATUS;
	char reqstatus = (char)status;
	char *reqbuf;
	size_t reqsize;
	char ansbuf[SHORT_ANSWER_BUFFER];
	size_t anssize = SHORT_ANSWER_BUFFER;
	
	if ((reqbuf = serialize_form_message((char *)&set_tcpuc_status_req, 0, &reqsize, reqid, reqstatus)) != NULL) {
		if ((streamconfig_request(sc, reqbuf, &reqsize, (char *)&ansbuf, &anssize)) && (ansbuf[1] == STREAMCONFIG_RES_SUCCESS)) {
			streamconfig_file_append_int(sc, "tcp_unicast_status", (gint)status);
			streamconfig_file_flush(sc);
			return TRUE;
		}
	}
	
	return FALSE;
}

gboolean streamconfig_set_connection_details(struct _streamconfig *sc, struct _streamconnparams *sendparams)
{
	const char set_connections_details_req[] = SET_CONNECTION_DETAILS_REQ;
	char reqid = STREAMCONFIG_SET_CONNECTION_DETAILS;
	char *reqbuf;
	size_t reqsize;
	char ansbuf[SHORT_ANSWER_BUFFER];
	size_t anssize = SHORT_ANSWER_BUFFER;
		
	if ((sc == NULL) || (sendparams == NULL)) return FALSE;
	
	if ((reqbuf = serialize_form_message((char *)&set_connections_details_req, 0, &reqsize, reqid, sendparams->udpbcint, sendparams->udpbcport, 
					sendparams->udpmcint, sendparams->udpmcgroup, sendparams->udpmcport, sendparams->udpmcloop, sendparams->udpmcttl,
					sendparams->tcpucport, sendparams->tcpucqueue, sendparams->tcpucclients)) != NULL) {
		if ((streamconfig_request(sc, reqbuf, &reqsize, (char *)&ansbuf, &anssize)) && (ansbuf[1] == STREAMCONFIG_RES_SUCCESS)) {
			if (sendparams->udpbcint[0] != '\0') 
				streamconfig_file_append_string(sc, "udp_broadcast_interface", (gchar *)sendparams->udpbcint);
			if (sendparams->udpmcint[0] != '\0') {
				streamconfig_file_append_string(sc, "udp_multicast_interface", (gchar *)sendparams->udpmcint);
				streamconfig_add_multicast_route(sc, (gchar *)&sendparams->udpmcint);
			}
			if (sendparams->udpmcgroup[0] != '\0') 
				streamconfig_file_append_string(sc, "udp_multicast_group", (gchar *)sendparams->udpmcgroup);
			streamconfig_file_append_int(sc, "udp_broadcast_port", (gint)sendparams->udpbcport);
			streamconfig_file_append_int(sc, "udp_multicast_port", (gint)sendparams->udpmcport);
			streamconfig_file_append_int(sc, "udp_multicast_loopback", (gint)sendparams->udpmcloop);
			streamconfig_file_append_int(sc, "udp_multicast_ttl", (gint)sendparams->udpmcttl);
			streamconfig_file_append_int(sc, "tcp_unicast_port", (gint)sendparams->tcpucport);
			streamconfig_file_append_int(sc, "tcp_unicast_queue", (gint)sendparams->tcpucqueue);
			streamconfig_file_append_int(sc, "tcp_unicast_clients", (gint)sendparams->tcpucclients);
			streamconfig_file_flush(sc);
			return TRUE;
		}
	}
	
	return FALSE;
}

void streamconfig_stream_content_foreach(unsigned int *params, unsigned int count, int seq, int size, char **buffer, void *data)
{
	GSList *mediastreams;
	struct _mediastream *mediastream;
	
	mediastreams = (GSList *)data;
	mediastream = (struct _mediastream *)g_slist_nth_data(mediastreams, seq);
	
	serialize_form_message_buffer(params, count, buffer, MSGOFFSET(seq, size), mediastream->type, mediastream->value, mediastream->enabled);
}

gboolean streamconfig_set_stream_data(struct _streamconfig *sc, gchar *svcname, GSList *mediastreams, guint16 caid, gchar *cakey)
{
	gint i;
	struct _mediastream *mediastream;
	gboolean result = FALSE;
	const char set_stream_data_req[] = SET_STREAM_DATA_REQ;
	char reqid = STREAMCONFIG_SET_STREAM_DATA;
	char *reqbuf;
	size_t reqsize;
	char ansbuf[SHORT_ANSWER_BUFFER];
	size_t anssize = SHORT_ANSWER_BUFFER;
	const char set_stream_content_array_req[] = SET_STREAM_CONTENT_ARRAY_REQ;
	guint numstreams;
	
	
	if ((sc == NULL) || (svcname == NULL)) return FALSE;
			
	if ((reqbuf = serialize_form_message((char *)&set_stream_data_req, 0, &reqsize, reqid, svcname, cakey, caid)) != NULL) {
		if ((streamconfig_request(sc, reqbuf, &reqsize, (char *)&ansbuf, &anssize)) && (ansbuf[1] == STREAMCONFIG_RES_SUCCESS)) {
			streamconfig_file_append_string(sc, "service_name", svcname);
			streamconfig_file_append_int(sc, "service_caid", caid); 
			streamconfig_file_append_string(sc, "service_cakey", cakey);
			streamconfig_file_flush(sc);
			result |= TRUE;
		}
	}
	
	free(reqbuf);
	
	if (mediastreams == NULL) return result;
	
	numstreams = g_slist_length(mediastreams);
	if (numstreams > 256) {
		numstreams = 256;
	} else if (numstreams == 0) {
		return result;
	}
	
	reqid = STREAMCONFIG_SET_STREAM_CONTENT;
	
	if ((reqbuf = serialize_form_array_message((char *)&set_stream_content_array_req, reqid, numstreams, &reqsize, streamconfig_stream_content_foreach, mediastreams)) != NULL) {
		if ((streamconfig_request(sc, reqbuf, &reqsize, (char *)&ansbuf, &anssize)) && (ansbuf[1] == STREAMCONFIG_RES_SUCCESS)) {
			result |= TRUE;
		}
	}
	
	free(reqbuf);
	
	return result;
}

gboolean streamconfig_set_device(struct _streamconfig *sc, gchar adapter, gchar dvr)
{
	const char set_device_req[] = SET_DEVICE_REQ;
	char reqid = STREAMCONFIG_SET_DEVICE;
	char *reqbuf;
	size_t reqsize;
	char ansbuf[SHORT_ANSWER_BUFFER];
	size_t anssize = SHORT_ANSWER_BUFFER;
	
	if (sc == NULL) return FALSE;
	
	if ((reqbuf = serialize_form_message((char *)&set_device_req, 0, &reqsize, reqid, adapter, dvr)) != NULL) {
		if ((streamconfig_request(sc, reqbuf, &reqsize, (char *)&ansbuf, &anssize)) && (ansbuf[1] == STREAMCONFIG_RES_SUCCESS)) {
			streamconfig_file_append_int(sc, "adapter", (gint)adapter);
			streamconfig_file_append_int(sc, "dvr", (gint)dvr);
			streamconfig_file_flush(sc);
			return TRUE;
		}
	}
	
	return FALSE;
}

void streamconfig_get_clients_callback(unsigned int *params, unsigned int count, int seq, int size, char **message, void *data)
{
	GSList *clients;
	struct _strclient *strclient;
	
	clients = *((GSList **)data);
	strclient = (struct _strclient *)g_malloc0(sizeof(struct _strclient));
		
	serialize_parse_message_buffer(params, count, message, MSGOFFSET(seq, size), &strclient->family, &strclient->type,
									&strclient->pid, &strclient->seq, &strclient->pidfilter, &strclient->address, &strclient->useragent,
									&strclient->starttime, &strclient->transmittedbytes);
	//printf("--> pid: %i, address: %s, ua: %s\n", strclient->pid, strclient->address, strclient->useragent);
	strclient->sticky = FALSE;
	*((GSList **)data) = g_slist_prepend(clients, strclient);
}

GSList *streamconfig_get_clients(struct _streamconfig *sc)
{
	const char get_simple_req[] = GET_SIMPLE_REQ;
	const char get_clients_array_req[] = GET_CLIENTS_ARRAY_REQ;
	char reqid;
	char *reqbuf;
	size_t reqsize;
	char ansbuf[4096];
	size_t anssize;
	char result;
	GSList *clients = NULL;
	
	if (sc == NULL) return NULL;
	
	reqid = STREAMCONFIG_GET_CLIENTS;
	anssize = 4096;
	
	if ((reqbuf = serialize_form_message((char *)&get_simple_req, 0, &reqsize, reqid)) != NULL) {
		if (streamconfig_request(sc, reqbuf, &reqsize, (char *)&ansbuf, &anssize)) {
			if (serialize_parse_array_message((char *)&get_clients_array_req, NULL, (char *)&ansbuf, NULL, streamconfig_get_clients_callback, &clients)) {
				clients = g_slist_reverse(clients);
				return clients;
			}
		}
	}
	
	return NULL;
}

void streamconfig_free_clients(GSList *clients)
{
	if (clients == NULL) return;
	
	g_slist_free(clients);
}

gboolean streamconfig_get_service_flags(struct _streamconfig *sc, gboolean *record, gboolean *udpbc, gboolean *udpmc, gboolean *tcpuc)
{
	const char get_simple_req[] = GET_SIMPLE_REQ;
	const char get_service_flags_req[] = GET_SERVICE_FLAGS_REQ;
	char reqid;
	char *reqbuf;
	size_t reqsize;
	char ansbuf[SHORT_ANSWER_BUFFER];
	size_t anssize;
	
	char result;
	char recordstatus;
	char udpbcstatus;
	char udpmcstatus;
	char tcpucstatus;
	
	
	if (sc == NULL) return FALSE;
		
	reqid = STREAMCONFIG_GET_SERVICE_FLAGS;
	anssize = SHORT_ANSWER_BUFFER;
	
	if ((reqbuf = serialize_form_message((char *)&get_simple_req, 0, &reqsize, reqid)) != NULL) {
		if ((streamconfig_request(sc, reqbuf, &reqsize, (char *)&ansbuf, &anssize)) && (ansbuf[1] == STREAMCONFIG_RES_SUCCESS)) {
			if (serialize_parse_message((char *)&get_service_flags_req, (char *)&ansbuf, 0, &reqid, &result, &recordstatus, &udpbcstatus, &udpmcstatus, &tcpucstatus) > 0) {
				if (record != NULL) *record = (gboolean)recordstatus;
				if (udpbc != NULL) *udpbc = (gboolean)udpbcstatus;
				if (udpmc != NULL) *udpmc = (gboolean)udpmcstatus;
				if (tcpuc != NULL) *tcpuc = (gboolean)tcpucstatus;
				return TRUE;
			}
		}
	}
	
	if (record != NULL) *record = (gboolean)streamconfig_file_read_int(sc, "record_status", 0);
	if (udpbc != NULL) *udpbc = (gboolean)streamconfig_file_read_int(sc, "udp_broadcast_status", 0);
	if (udpmc != NULL) *udpmc = (gboolean)streamconfig_file_read_int(sc, "udp_multicast_status", 0);
	if (tcpuc != NULL) *tcpuc = (gboolean)streamconfig_file_read_int(sc, "tcp_unicast_status", 0);
	return FALSE;
}

gboolean streamconfig_get_connection_details(struct _streamconfig *sc, struct _streamconnparams *recvparams)
{
	const char get_simple_req[] = GET_SIMPLE_REQ;
	const char get_connection_details_req[] = GET_CONNECTION_DETAILS_REQ;
	char reqid = STREAMCONFIG_GET_CONNECTION_DETAILS;
	char *reqbuf;
	size_t reqsize;
	char ansbuf[ANSWER_BUFFER];
	size_t anssize = ANSWER_BUFFER;
	char result;
	gchar *str;
	gsize length;
	
	
	if ((sc == NULL) || (recvparams == NULL)) return FALSE;
	
	memset(recvparams, 0, sizeof(struct _streamconnparams));
		
	if ((reqbuf = serialize_form_message((char *)&get_simple_req, 0, &reqsize, reqid)) != NULL) {
		if ((streamconfig_request(sc, reqbuf, &reqsize, (char *)&ansbuf, &anssize)) && (ansbuf[1] == STREAMCONFIG_RES_SUCCESS)) {
			if (serialize_parse_message_static((char *)&get_connection_details_req, (char *)&ansbuf, 0, &reqid, &result, &recvparams->udpbcint, &recvparams->udpbcport, 
					&recvparams->udpmcint, &recvparams->udpmcgroup, &recvparams->udpmcport, &recvparams->udpmcloop, &recvparams->udpmcttl,
					&recvparams->tcpucport, &recvparams->tcpucqueue, &recvparams->tcpucclients) > 0) {
				return TRUE;
			}
		}
	}
	
	str = streamconfig_file_read_string(sc, "udp_broadcast_interface", "lo");
	if ((length = strlen(str)) > IFNAMSIZ) {
		strncpy(recvparams->udpbcint, str, IFNAMSIZ);
		recvparams->udpbcint[IFNAMSIZ] = '\0';
	} else {
		strncpy(recvparams->udpbcint, str, length);
		recvparams->udpbcint[length] = '\0';
	}
	
	
	str = streamconfig_file_read_string(sc, "udp_multicast_interface", "lo");
	if ((length = strlen(str)) > IFNAMSIZ) {
		strncpy(recvparams->udpmcint, str, IFNAMSIZ);
		recvparams->udpmcint[IFNAMSIZ] = '\0';
	} else {
		strncpy(recvparams->udpmcint, str, length);
		recvparams->udpmcint[length] = '\0';
	}
	
	str = streamconfig_file_read_string(sc, "udp_multicast_group", "225.0.0.0");
	if ((length = strlen(str)) > INET_ADDRSTRLEN) {
		strncpy(recvparams->udpmcgroup, str, INET_ADDRSTRLEN);
		recvparams->udpmcgroup[INET_ADDRSTRLEN] = '\0';
	} else {
		strncpy(recvparams->udpmcgroup, str, length);
		recvparams->udpmcgroup[length] = '\0';
	}
	
	recvparams->udpbcport = streamconfig_file_read_int(sc, "udp_broadcast_port", 4000);
	recvparams->udpmcport = streamconfig_file_read_int(sc, "udp_multicast_port", 4000);
	recvparams->udpmcloop = streamconfig_file_read_int(sc, "udp_multicast_loopback", 1);
	recvparams->udpmcttl = streamconfig_file_read_int(sc, "udp_multicast_ttl", 1);
	recvparams->tcpucport = streamconfig_file_read_int(sc, "tcp_unicast_port", 4000);
	recvparams->tcpucqueue = streamconfig_file_read_int(sc, "tcp_unicast_queue", 10);
	recvparams->tcpucclients = streamconfig_file_read_int(sc, "tcp_unicast_clients", 20);
	
	return TRUE;
}


gboolean streamconfig_stop_client(struct _streamconfig *sc, guint clientid)
{
	const char stop_client_req[] = STOP_CLIENT_REQ;
	char reqid = STREAMCONFIG_STOP_CLIENT;
	char *reqbuf;
	size_t reqsize;
	char ansbuf[SHORT_ANSWER_BUFFER];
	size_t anssize = SHORT_ANSWER_BUFFER;
	
	if ((reqbuf = serialize_form_message((char *)&stop_client_req, 0, &reqsize, reqid, clientid)) != NULL) {
		if ((streamconfig_request(sc, reqbuf, &reqsize, (char *)&ansbuf, &anssize)) && (ansbuf[1] == STREAMCONFIG_RES_SUCCESS)) {
			return TRUE;
		}
	}
	
	return FALSE;
}


static gboolean streamconfig_request(struct _streamconfig *sc, char *reqbuf, size_t *reqsize, char *ansbuf, size_t *anssize)
{
	gint ready;
	fd_set readset;
	struct sockaddr_un server;
	struct timeval tv;
	static struct msghdr msgsend, msgrecv;
	struct iovec iovsend[1], iovrecv[1];
	gsize bytes;
	gchar request;
		
	
	if (!streamconfig_get_running(sc)) return FALSE;
	
	request = reqbuf[0];
	
	msgsend.msg_name = &sc->serveraddr;
	msgsend.msg_namelen = sizeof(struct sockaddr_un);
	msgsend.msg_iov = iovsend;
	msgsend.msg_iovlen = 1;
	iovsend[0].iov_base = reqbuf;
	iovsend[0].iov_len = *reqsize;
	
	bytes = sendmsg(sc->clientsock, &msgsend, 0);
	
	//printf("R--> 0x%x::0x%x\n", (unsigned char)reqbuf[0], (unsigned char)reqbuf[1]);
	
	while (1) {
		FD_ZERO(&readset);
		FD_SET(sc->clientsock, &readset);
		
		tv.tv_sec = 0;
		tv.tv_usec = 50000;
		
		ready = select(sc->clientsock+1, &readset, NULL, NULL, &tv);
		
		if ((ready) && (FD_ISSET(sc->clientsock, &readset))) {
			msgrecv.msg_name = &server;
			msgrecv.msg_namelen = sizeof(struct sockaddr_un);
			msgrecv.msg_iov = iovrecv;
			msgrecv.msg_iovlen = 1;
			iovrecv[0].iov_base = ansbuf;
			iovrecv[0].iov_len = *anssize;
			
			memset(ansbuf, 0, *anssize);
					
			bytes = recvmsg(sc->clientsock, &msgrecv, 0);
			
			//printf("A--> 0x%x::0x%x\n", (unsigned char)ansbuf[0], (unsigned char)ansbuf[1]);
			
			if (ansbuf[0] != request) continue; else return TRUE;
		} else return FALSE;
	}
	
	return FALSE;
}
